@@ -1,7 +1,7 @@
 %%
 %% %CopyrightBegin%
 %%
-%% Copyright Ericsson AB 2007-2019. All Rights Reserved.
+%% Copyright Ericsson AB 2007-2020. All Rights Reserved.
 %%
 %% Licensed under the Apache License, Version 2.0 (the "License");
 %% you may not use this file except in compliance with the License.
@@ -88,7 +88,7 @@
 %% gen_statem callbacks
 -export([callback_mode/0, terminate/3, code_change/4, format_status/2]).
  
--export([encode_handshake/4]).
+-export([encode_handshake/4, send_key_update/2, update_cipher_key/2]).
 
 -define(DIST_CNTRL_SPAWN_OPTS, [{priority, max}]).
 
@@ -98,28 +98,28 @@
 %%====================================================================
 %% Setup
 %%====================================================================
-start_fsm(Role, Host, Port, Socket, {#{erl_dist := false},_, Tracker} = Opts,
+start_fsm(Role, Host, Port, Socket, {#{erl_dist := false},_, Trackers} = Opts,
 	  User, {CbModule, _,_, _, _} = CbInfo, 
 	  Timeout) -> 
     try 
         {ok, Sender} = tls_sender:start(),
 	{ok, Pid} = tls_connection_sup:start_child([Role, Sender, Host, Port, Socket, 
 						    Opts, User, CbInfo]), 
-	{ok, SslSocket} = ssl_connection:socket_control(?MODULE, Socket, [Pid, Sender], CbModule, Tracker),
+	{ok, SslSocket} = ssl_connection:socket_control(?MODULE, Socket, [Pid, Sender], CbModule, Trackers),
         ssl_connection:handshake(SslSocket, Timeout)
     catch
 	error:{badmatch, {error, _} = Error} ->
 	    Error
     end;
 
-start_fsm(Role, Host, Port, Socket, {#{erl_dist := true},_, Tracker} = Opts,
+start_fsm(Role, Host, Port, Socket, {#{erl_dist := true},_, Trackers} = Opts,
 	  User, {CbModule, _,_, _, _} = CbInfo, 
 	  Timeout) -> 
     try 
         {ok, Sender} = tls_sender:start([{spawn_opt, ?DIST_CNTRL_SPAWN_OPTS}]),
 	{ok, Pid} = tls_connection_sup:start_child_dist([Role, Sender, Host, Port, Socket, 
 							 Opts, User, CbInfo]), 
-	{ok, SslSocket} = ssl_connection:socket_control(?MODULE, Socket, [Pid, Sender], CbModule, Tracker),
+	{ok, SslSocket} = ssl_connection:socket_control(?MODULE, Socket, [Pid, Sender], CbModule, Trackers),
         ssl_connection:handshake(SslSocket, Timeout)
     catch
 	error:{badmatch, {error, _} = Error} ->
@@ -221,7 +221,7 @@ activate_socket(#state{protocol_specific = #{active_n_toggle := true, active_n :
 next_record(State, CipherTexts, ConnectionStates, Check) ->
     next_record(State, CipherTexts, ConnectionStates, Check, []).
 %%
-next_record(#state{connection_env = #connection_env{negotiated_version = Version}} = State,
+next_record(#state{connection_env = #connection_env{negotiated_version = {3,4} = Version}} = State,
             [CT|CipherTexts], ConnectionStates0, Check, Acc) ->
     case tls_record:decode_cipher_text(Version, CT, ConnectionStates0, Check) of
         {#ssl_tls{type = ?APPLICATION_DATA, fragment = Fragment}, ConnectionStates} ->
@@ -242,10 +242,41 @@ next_record(#state{connection_env = #connection_env{negotiated_version = Version
             %% Not ?APPLICATION_DATA but we have accumulated fragments
             %% -> build an ?APPLICATION_DATA record with concatenated fragments
             %%    and forget about decrypting this record - we'll decrypt it again next time
+            %% Will not work for stream ciphers
             next_record_done(State, [CT|CipherTexts], ConnectionStates0,
                              #ssl_tls{type = ?APPLICATION_DATA, fragment = iolist_to_binary(lists:reverse(Acc))});
         #alert{} = Alert ->
             Alert
+    end;
+next_record(#state{connection_env = #connection_env{negotiated_version = Version}} = State,
+            [#ssl_tls{type = ?APPLICATION_DATA} = CT |CipherTexts], ConnectionStates0, Check, Acc) ->
+    case tls_record:decode_cipher_text(Version, CT, ConnectionStates0, Check) of
+        {#ssl_tls{type = ?APPLICATION_DATA, fragment = Fragment}, ConnectionStates} ->
+            case CipherTexts of
+                [] ->
+                    %% End of cipher texts - build and deliver an ?APPLICATION_DATA record
+                    %% from the accumulated fragments
+                    next_record_done(State, [], ConnectionStates,
+                                     #ssl_tls{type = ?APPLICATION_DATA,
+                                              fragment = iolist_to_binary(lists:reverse(Acc, [Fragment]))});
+                [_|_] ->
+                    next_record(State, CipherTexts, ConnectionStates, Check, [Fragment|Acc])
+            end;
+        #alert{} = Alert ->
+            Alert
+    end;
+next_record(State, CipherTexts, ConnectionStates, _, [_|_] = Acc) ->
+    next_record_done(State, CipherTexts, ConnectionStates,
+                     #ssl_tls{type = ?APPLICATION_DATA,
+                              fragment = iolist_to_binary(lists:reverse(Acc))});
+next_record(#state{connection_env = #connection_env{negotiated_version = Version}} = State,
+            [CT|CipherTexts], ConnectionStates0, Check, []) ->
+    case tls_record:decode_cipher_text(Version, CT, ConnectionStates0, Check) of      
+        {Record, ConnectionStates} ->
+            %% Singelton non-?APPLICATION_DATA record - deliver
+            next_record_done(State, CipherTexts, ConnectionStates, Record);
+        #alert{} = Alert ->
+            Alert
     end.
 
 next_record_done(#state{protocol_buffers = Buffers} = State, CipherTexts, ConnectionStates, Record) ->
@@ -256,12 +287,15 @@ next_record_done(#state{protocol_buffers = Buffers} = State, CipherTexts, Connec
 next_event(StateName, Record, State) ->
     next_event(StateName, Record, State, []).
 %%
-next_event(StateName, no_record, State0, Actions) ->
+next_event(StateName, no_record, #state{static_env = #static_env{role = Role}} = State0, Actions) ->
     case next_record(StateName, State0) of
  	{no_record, State} ->
             ssl_connection:hibernate_after(StateName, State, Actions);
         {Record, State} ->
-            next_event(StateName, Record, State, Actions)
+            next_event(StateName, Record, State, Actions);
+        #alert{} = Alert ->
+            ssl_connection:handle_normal_shutdown(Alert#alert{role = Role}, StateName, State0),
+	    {stop, {shutdown, own_alert}, State0}
     end;
 next_event(StateName,  #ssl_tls{} = Record, State, Actions) ->
     {next_state, StateName, State, [{next_event, internal, {protocol_record, Record}} | Actions]};
@@ -275,32 +309,33 @@ handle_protocol_record(#ssl_tls{type = ?APPLICATION_DATA, fragment = Data}, Stat
     case ssl_connection:read_application_data(Data, State0) of
        {stop, _, _} = Stop->
             Stop;
-       {Record, #state{start_or_recv_from = Caller} = State1} ->
+        {Record, #state{start_or_recv_from = Caller} = State} ->
             TimerAction = case Caller of
                               undefined -> %% Passive recv complete cancel timer
                                   [{{timeout, recv}, infinity, timeout}];
                               _ ->
                                   []
                           end,
-            {next_state, StateName, State, Actions} = next_event(StateName, Record, State1, TimerAction), 
-            ssl_connection:hibernate_after(StateName, State, Actions)
+            next_event(StateName, Record, State, TimerAction)
     end;
 handle_protocol_record(#ssl_tls{type = ?APPLICATION_DATA, fragment = Data}, StateName, State0) ->
     case ssl_connection:read_application_data(Data, State0) of
 	{stop, _, _} = Stop->
             Stop;
-	{Record, State1} ->
-            {next_state, StateName, State, Actions} = next_event(StateName, Record, State1), 
-            ssl_connection:hibernate_after(StateName, State, Actions)
+	{Record, State} ->
+            next_event(StateName, Record, State)
     end;
 %%% TLS record protocol level handshake messages 
 handle_protocol_record(#ssl_tls{type = ?HANDSHAKE, fragment = Data}, 
 		    StateName, #state{protocol_buffers =
 					  #protocol_buffers{tls_handshake_buffer = Buf0} = Buffers,
                                       connection_env = #connection_env{negotiated_version = Version},
+                                      static_env = #static_env{role = Role},
 				      ssl_options = Options} = State0) ->
     try
-	EffectiveVersion = effective_version(Version, Options),
+	%% Calculate the effective version that should be used when decoding an incoming handshake
+	%% message.
+	EffectiveVersion = effective_version(Version, Options, Role),
 	{Packets, Buf} = tls_handshake:get_tls_handshake(EffectiveVersion,Data,Buf0, Options),
 	State =
 	    State0#state{protocol_buffers =
@@ -381,7 +416,6 @@ renegotiate(#state{static_env = #static_env{role = server,
 send_handshake(Handshake, State) ->
     send_handshake_flight(queue_handshake(Handshake, State)).
 
-
 queue_handshake(Handshake, #state{handshake_env = #handshake_env{tls_handshake_history = Hist0} = HsEnv,
 				  connection_env = #connection_env{negotiated_version = Version},
                                   flight_buffer = Flight0,
@@ -396,7 +430,10 @@ queue_handshake(Handshake, #state{handshake_env = #handshake_env{tls_handshake_h
                  handshake_env = HsEnv#handshake_env{tls_handshake_history = Hist},
 		 flight_buffer = Flight0 ++ [BinHandshake]}.
 
-
+-spec send_handshake_flight(StateIn) -> {StateOut, FlightBuffer} when
+      StateIn :: #state{},
+      StateOut :: #state{},
+      FlightBuffer :: list().
 send_handshake_flight(#state{static_env = #static_env{socket = Socket,
                                                       transport_cb = Transport},
 			     flight_buffer = Flight} = State0) ->
@@ -506,7 +543,7 @@ close(downgrade, _,_,_,_) ->
     ok;
 %% Other
 close(_, Socket, Transport, _,_) -> 
-    Transport:close(Socket).
+    tls_socket:close(Transport, Socket).
 protocol_name() ->
     "TLS".
 
@@ -514,8 +551,8 @@ protocol_name() ->
 %% Data handling
 %%====================================================================	     
 
-socket(Pids,  Transport, Socket, Tracker) ->
-    tls_socket:socket(Pids, Transport, Socket, ?MODULE, Tracker).
+socket(Pids,  Transport, Socket, Trackers) ->
+    tls_socket:socket(Pids, Transport, Socket, ?MODULE, Trackers).
 
 setopts(Transport, Socket, Other) ->
     tls_socket:setopts(Transport, Socket, Other).
@@ -543,26 +580,51 @@ init({call, From}, {start, Timeout},
             handshake_env = #handshake_env{renegotiation = {Renegotiation, _}} = HsEnv,
             connection_env = CEnv,
 	    ssl_options = #{log_level := LogLevel,
-                            versions := Versions} = SslOpts,
-	    session = #session{own_certificate = Cert} = Session0,
+                            %% Use highest version in initial ClientHello.
+                            %% Versions is a descending list of supported versions.
+                            versions := [HelloVersion|_] = Versions,
+                            session_tickets := SessionTickets} = SslOpts,
+	    session = NewSession,
 	    connection_states = ConnectionStates0
 	   } = State0) ->
     KeyShare = maybe_generate_client_shares(SslOpts),
+    Session = ssl_session:client_select_session({Host, Port, SslOpts}, Cache, CacheCb, NewSession),
+    %% Update UseTicket in case of automatic session resumption
+    {UseTicket, State1} = tls_handshake_1_3:maybe_automatic_session_resumption(State0),
+    TicketData = tls_handshake_1_3:get_ticket_data(self(), SessionTickets, UseTicket),
     Hello = tls_handshake:client_hello(Host, Port, ConnectionStates0, SslOpts,
-				       Cache, CacheCb, Renegotiation, Cert, KeyShare),
+                                       Session#session.session_id,
+                                       Renegotiation,
+                                       Session#session.own_certificate,
+                                       KeyShare,
+                                       TicketData),
 
-    HelloVersion = tls_record:hello_version(Versions),
     Handshake0 = ssl_handshake:init_handshake_history(),
+
+    %% Update pre_shared_key extension with binders (TLS 1.3)
+    Hello1 = tls_handshake_1_3:maybe_add_binders(Hello, TicketData, HelloVersion),
+
+    MaxFragEnum = maps:get(max_frag_enum, Hello1#client_hello.extensions, undefined),
+    ConnectionStates1 = ssl_record:set_max_fragment_length(MaxFragEnum, ConnectionStates0),
+
     {BinMsg, ConnectionStates, Handshake} =
-        encode_handshake(Hello,  HelloVersion, ConnectionStates0, Handshake0),
+        encode_handshake(Hello1,  HelloVersion, ConnectionStates1, Handshake0),
+
     tls_socket:send(Transport, Socket, BinMsg),
-    ssl_logger:debug(LogLevel, outbound, 'handshake', Hello),
+    ssl_logger:debug(LogLevel, outbound, 'handshake', Hello1),
     ssl_logger:debug(LogLevel, outbound, 'record', BinMsg),
 
-    State = State0#state{connection_states = ConnectionStates,
-                         connection_env = CEnv#connection_env{negotiated_version = HelloVersion}, %% Requested version
-                         session =
-                             Session0#session{session_id = Hello#client_hello.session_id},
+    %% RequestedVersion is used as the legacy record protocol version and shall be
+    %% {3,3} in case of TLS 1.2 and higher. In all other cases it defaults to the
+    %% lowest supported protocol version.
+    %%
+    %% negotiated_version is also used by the TLS 1.3 state machine and is set after
+    %% ServerHello is processed.
+    RequestedVersion = tls_record:hello_version(Versions),
+    State = State1#state{connection_states = ConnectionStates,
+                         connection_env = CEnv#connection_env{
+                                            negotiated_version = RequestedVersion},
+                         session = Session,
                          handshake_env = HsEnv#handshake_env{tls_handshake_history = Handshake},
                          start_or_recv_from = From,
                          key_share = KeyShare},
@@ -599,7 +661,7 @@ hello(internal, #client_hello{extensions = Extensions} = Hello,
     {next_state, user_hello, State#state{start_or_recv_from = undefined,
                                          handshake_env = HsEnv#handshake_env{hello = Hello}},
      [{reply, From, {ok, Extensions}}]};
-hello(internal, #server_hello{extensions = Extensions} = Hello, 
+hello(internal, #server_hello{extensions = Extensions} = Hello,
       #state{ssl_options = #{handshake := hello},
              handshake_env = HsEnv,
              start_or_recv_from = From} = State) ->
@@ -659,8 +721,9 @@ hello(internal, #server_hello{} = Hello,
              connection_env = #connection_env{negotiated_version = ReqVersion} = CEnv,
 	     static_env = #static_env{role = client},
              handshake_env = #handshake_env{renegotiation = {Renegotiation, _}},
+             session = #session{session_id = OldId},
 	     ssl_options = SslOptions} = State) ->
-    case tls_handshake:hello(Hello, SslOptions, ConnectionStates0, Renegotiation) of
+    case tls_handshake:hello(Hello, SslOptions, ConnectionStates0, Renegotiation, OldId) of
 	#alert{} = Alert -> %%TODO
 	    ssl_connection:handle_own_alert(Alert, ReqVersion, hello,
                                             State#state{connection_env =
@@ -670,12 +733,15 @@ hello(internal, #server_hello{} = Hello,
 	    ssl_connection:handle_session(Hello, 
 					  Version, NewId, ConnectionStates, ProtoExt, Protocol, State);
         %% TLS 1.3
-        {next_state, wait_sh} ->
+        {next_state, wait_sh, SelectedVersion} ->
             %% Continue in TLS 1.3 'wait_sh' state
-            {next_state, wait_sh, State, [{next_event, internal, Hello}]}
+            {next_state, wait_sh,
+             State#state{
+               connection_env = CEnv#connection_env{negotiated_version = SelectedVersion}},
+             [{next_event, internal, Hello}]}
     end;
 hello(info, Event, State) ->
-    gen_info(Event, ?FUNCTION_NAME, State);
+    handle_info(Event, ?FUNCTION_NAME, State);
 hello(Type, Event, State) ->
     gen_handshake(?FUNCTION_NAME, Type, Event, State).
 
@@ -722,15 +788,19 @@ connection({call, From}, {user_renegotiate, WriteState},
      [{next_event,{call, From}, renegotiate}]};
 connection({call, From}, 
            {close, {Pid, _Timeout}}, 
-           #state{connection_env = #connection_env{terminated = closed} =CEnv} = State) ->
+           #state{connection_env = #connection_env{terminated = closed} = CEnv,
+                 protocol_specific = PS} = State) ->
     {next_state, downgrade, State#state{connection_env = 
-                                            CEnv#connection_env{terminated = true, 
-                                                                downgrade = {Pid, From}}}, 
+                                            CEnv#connection_env{terminated = true,
+                                                                downgrade = {Pid, From}},
+                                        protocol_specific = PS#{active_n_toggle => true,
+                                                                active_n => 1}
+                                       },
      [{next_event, internal, ?ALERT_REC(?WARNING, ?CLOSE_NOTIFY)}]};
 connection({call, From}, 
            {close,{Pid, Timeout}},
            #state{connection_states = ConnectionStates,
-                  protocol_specific = #{sender := Sender},
+                  protocol_specific = #{sender := Sender} = PS,
                   connection_env = CEnv
                  } = State0) ->
     case tls_sender:downgrade(Sender, Timeout) of
@@ -744,7 +814,10 @@ connection({call, From},
                                                 ConnectionStates#{current_write => Write}}),
             {next_state, downgrade, State#state{connection_env = 
                                                     CEnv#connection_env{downgrade = {Pid, From},
-                                                                        terminated = true}}, 
+                                                                        terminated = true},
+                                                protocol_specific = PS#{active_n_toggle => true,
+                                                                        active_n => 1}
+                                               },
              [{timeout, Timeout, downgrade}]};
         {error, timeout} ->
             {stop_and_reply, {shutdown, downgrade_fail}, [{reply, From, {error, timeout}}]}
@@ -762,11 +835,14 @@ connection(internal, #hello_request{},
 		  connection_states = ConnectionStates} = State0) ->
     try tls_sender:peer_renegotiate(Pid) of
         {ok, Write} ->
+            Session = ssl_session:client_select_session({Host, Port, SslOpts}, Cache, CacheCb, Session0),
             Hello = tls_handshake:client_hello(Host, Port, ConnectionStates, SslOpts,
-                                               Cache, CacheCb, Renegotiation, Cert, undefined),
-            {State, Actions} = send_handshake(Hello, State0#state{connection_states = ConnectionStates#{current_write => Write}}),
-            next_event(hello, no_record, State#state{session = Session0#session{session_id
-                                                                      = Hello#client_hello.session_id}}, Actions)
+                                               Session#session.session_id,
+                                               Renegotiation, Cert, undefined,
+                                               undefined),
+            {State, Actions} = send_handshake(Hello, State0#state{connection_states = ConnectionStates#{current_write => Write},
+                                                                  session = Session}),
+            next_event(hello, no_record, State, Actions)
         catch 
             _:_ ->
                 {stop, {shutdown, sender_blocked}, State0}
@@ -774,19 +850,17 @@ connection(internal, #hello_request{},
 connection(internal, #hello_request{},
 	   #state{static_env = #static_env{role = client,
                                            host = Host,
-                                           port = Port,
-                                           session_cache = Cache,
-                                           session_cache_cb = CacheCb},
+                                           port = Port},
                   handshake_env = #handshake_env{renegotiation = {Renegotiation, _}},
-		  session = #session{own_certificate = Cert} = Session0,
+		  session = #session{own_certificate = Cert},
 		  ssl_options = SslOpts, 
 		  connection_states = ConnectionStates} = State0) ->
     Hello = tls_handshake:client_hello(Host, Port, ConnectionStates, SslOpts,
-				       Cache, CacheCb, Renegotiation, Cert, undefined),
+                                       <<>>, Renegotiation, Cert, undefined,
+                                       undefined),
 
     {State, Actions} = send_handshake(Hello, State0),
-    next_event(hello, no_record, State#state{session = Session0#session{session_id
-                                                                        = Hello#client_hello.session_id}}, Actions);
+    next_event(hello, no_record, State, Actions);
 connection(internal, #client_hello{} = Hello, 
 	   #state{static_env = #static_env{role = server},
                   handshake_env = #handshake_env{allow_renegotiate = true}= HsEnv,
@@ -813,11 +887,21 @@ connection(internal, #client_hello{},
     State = reinit_handshake_data(State0),
     next_event(?FUNCTION_NAME, no_record, State);
 
-connection(internal, #new_session_ticket{}, State) ->
+connection(internal, #new_session_ticket{} = NewSessionTicket, State) ->
     %% TLS 1.3
-    %% Drop NewSessionTicket (currently not supported)
+    handle_new_session_ticket(NewSessionTicket, State),
     next_event(?FUNCTION_NAME, no_record, State);
 
+connection(internal, #key_update{} = KeyUpdate, State0) ->
+    %% TLS 1.3
+    case handle_key_update(KeyUpdate, State0) of
+        {ok, State} ->
+            next_event(?FUNCTION_NAME, no_record, State);
+        {error, State, Alert} ->
+            ssl_connection:handle_own_alert(Alert, {3,4}, connection, State),
+            next_event(?FUNCTION_NAME, no_record, State)
+    end;
+
 connection(Type, Event, State) ->
     ssl_connection:?FUNCTION_NAME(Type, Event, State, ?MODULE).
 
@@ -981,7 +1065,7 @@ code_change(_OldVsn, StateName, State, _) ->
 %%--------------------------------------------------------------------
 %%% Internal functions
 %%--------------------------------------------------------------------
-initial_state(Role, Sender, Host, Port, Socket, {SSLOptions, SocketOptions, Tracker}, User,
+initial_state(Role, Sender, Host, Port, Socket, {SSLOptions, SocketOptions, Trackers}, User,
 	      {CbModule, DataTag, CloseTag, ErrorTag, PassiveTag}) ->
     #{beast_mitigation := BeastMitigation,
       erl_dist := IsErlDist,
@@ -1012,8 +1096,8 @@ initial_state(Role, Sender, Host, Port, Socket, {SSLOptions, SocketOptions, Trac
                      port = Port,
                      socket = Socket,
                      session_cache_cb = SessionCacheCb,
-                     tracker = Tracker
-                    },
+                     trackers = Trackers
+                    },  
     #state{
        static_env = InitStatEnv,
        handshake_env = #handshake_env{
@@ -1040,11 +1124,12 @@ initialize_tls_sender(#state{static_env = #static_env{
                                              role = Role,
                                              transport_cb = Transport,
                                              socket = Socket,
-                                             tracker = Tracker
+                                             trackers = Trackers
                                             },
                              connection_env = #connection_env{negotiated_version = Version},
                              socket_options = SockOpts, 
                              ssl_options = #{renegotiate_at := RenegotiateAt,
+                                             key_update_at := KeyUpdateAt,
                                              log_level := LogLevel},
                              connection_states = #{current_write := ConnectionWriteState},
                              protocol_specific = #{sender := Sender}}) ->
@@ -1052,10 +1137,11 @@ initialize_tls_sender(#state{static_env = #static_env{
              role => Role,
              socket => Socket,
              socket_options => SockOpts,
-             tracker => Tracker,
+             trackers => Trackers,
              transport_cb => Transport,
              negotiated_version => Version,
              renegotiate_at => RenegotiateAt,
+             key_update_at => KeyUpdateAt,
              log_level => LogLevel},
     tls_sender:initialize(Sender, Init).
 
@@ -1065,17 +1151,25 @@ next_tls_record(Data, StateName,
                                                       tls_cipher_texts = CT0} = Buffers,
                                 ssl_options = SslOpts} = State0) ->
     Versions =
-        %% TLS 1.3 Client/Server
-        %% - Ignore TLSPlaintext.legacy_record_version
-        %% - Verify that TLSCiphertext.legacy_record_version is set to  0x0303 for all records
-        %%   other than an initial ClientHello, where it MAY also be 0x0301.
+        %% TLSPlaintext.legacy_record_version is ignored in TLS 1.3 and thus all
+        %% record version are accepted when receiving initial ClientHello and
+        %% ServerHello. This can happen in state 'hello' in case of all TLS
+        %% versions and also in state 'start' when TLS 1.3 is negotiated.
+        %% After the version is negotiated all subsequent TLS records shall have
+        %% the proper legacy_record_version (= negotiated_version).
+        %% Note: TLS record version {3,4} is used internally in TLS 1.3 and at this
+        %% point it is the same as the negotiated protocol version.
+        %% TODO: Refactor state machine and introduce a record_protocol_version beside
+        %% the negotiated_version.
         case StateName of
-            hello ->
+            State when State =:= hello orelse
+                       State =:= start ->
                 [tls_record:protocol_version(Vsn) || Vsn <- ?ALL_AVAILABLE_VERSIONS];
             _ ->
                 State0#state.connection_env#connection_env.negotiated_version
         end,
-    case tls_record:get_tls_records(Data, Versions, Buf0, SslOpts) of
+    #{current_write := #{max_fragment_length := MaxFragLen}} = State0#state.connection_states,
+    case tls_record:get_tls_records(Data, Versions, Buf0, MaxFragLen, SslOpts) of
 	{Records, Buf1} ->
 	    CT1 = CT0 ++ Records,
 	    next_record(StateName, State0#state{protocol_buffers =
@@ -1096,24 +1190,48 @@ tls_handshake_events(Packets) ->
 
 %% raw data from socket, upack records
 handle_info({Protocol, _, Data}, StateName,
-            #state{static_env = #static_env{data_tag = Protocol}} = State0) ->
+            #state{static_env = #static_env{data_tag = Protocol},
+                   connection_env = #connection_env{negotiated_version = Version}} = State0) ->
     case next_tls_record(Data, StateName, State0) of
 	{Record, State} ->
 	    next_event(StateName, Record, State);
 	#alert{} = Alert ->
-	    ssl_connection:handle_normal_shutdown(Alert, StateName, State0), 
-	    {stop, {shutdown, own_alert}, State0}
+	    ssl_connection:handle_own_alert(Alert, Version, StateName, State0)
     end;
 handle_info({PassiveTag, Socket},  StateName, 
             #state{static_env = #static_env{socket = Socket,
                                             passive_tag = PassiveTag},
+                   start_or_recv_from = From,
+                   protocol_buffers = #protocol_buffers{tls_cipher_texts = CTs},
                    protocol_specific = PS
-                  } = State) ->
-    next_event(StateName, no_record, 
-               State#state{protocol_specific = PS#{active_n_toggle => true}});
+                  } = State0) ->
+    case (From =/= undefined) andalso (CTs == []) of
+        true ->
+            {Record, State} = activate_socket(State0#state{protocol_specific = PS#{active_n_toggle => true}}),
+            next_event(StateName, Record, State);
+        false ->
+            next_event(StateName, no_record, 
+                       State0#state{protocol_specific = PS#{active_n_toggle => true}})
+    end;
 handle_info({CloseTag, Socket}, StateName,
-            #state{static_env = #static_env{socket = Socket, close_tag = CloseTag},
+            #state{static_env = #static_env{
+                                   role = Role,
+                                   host = Host,
+                                   port = Port,
+                                   socket = Socket, 
+                                   close_tag = CloseTag},
+                   handshake_env = #handshake_env{renegotiation = Type},
                    connection_env = #connection_env{negotiated_version = Version},
+                   session = Session} = State) when  StateName =/= connection ->
+    ssl_connection:maybe_invalidate_session(Version, Type, Role, Host, Port, Session),
+    Alert = ?ALERT_REC(?FATAL, ?CLOSE_NOTIFY, transport_closed),
+    ssl_connection:handle_normal_shutdown(Alert#alert{role = Role}, StateName, State),
+    {stop, {shutdown, transport_closed}, State};
+handle_info({CloseTag, Socket}, StateName,
+            #state{static_env = #static_env{
+                                   role = Role,
+                                   socket = Socket,
+                                   close_tag = CloseTag},
                    socket_options = #socket_options{active = Active},
                    protocol_buffers = #protocol_buffers{tls_cipher_texts = CTs},
                    user_data_buffer = {_,BufferSize,_},
@@ -1126,18 +1244,18 @@ handle_info({CloseTag, Socket}, StateName,
 
     case (Active == false) andalso ((CTs =/= []) or (BufferSize =/= 0)) of
         false ->
-            case Version of
-                {1, N} when N >= 1 ->
-                    ok;
-                _ ->
-                    %% As invalidate_sessions here causes performance issues,
-                    %% we will conform to the widespread implementation
-                    %% practice and go aginst the spec
-                    %%invalidate_session(Role, Host, Port, Session)
-                    ok
-            end,
-
-            ssl_connection:handle_normal_shutdown(?ALERT_REC(?FATAL, ?CLOSE_NOTIFY), StateName, State),
+            %% As invalidate_sessions here causes performance issues,
+            %% we will conform to the widespread implementation
+            %% practice and go aginst the spec
+            %% case Version of
+            %%     {3, N} when N >= 1 ->
+            %%         ok;
+            %%     _ ->
+            %%         invalidate_session(Role, Host, Port, Session)
+            %%         ok
+            %% end,
+            Alert = ?ALERT_REC(?FATAL, ?CLOSE_NOTIFY, transport_closed),
+            ssl_connection:handle_normal_shutdown(Alert#alert{role = Role}, StateName, State),
             {stop, {shutdown, transport_closed}, State};
         true ->
             %% Fixes non-delivery of final TLS record in {active, once}.
@@ -1323,7 +1441,99 @@ choose_tls_version(_, _) ->
     'tls_v1.2'.
 
 
-effective_version(undefined, #{versions := [Version|_]}) ->
+%% Special version handling for TLS 1.3 clients:
+%% In the shared state 'init' negotiated_version is set to requested version and
+%% that is expected by the legacy part of the state machine. However, in order to
+%% be able to process new TLS 1.3 extensions, the effective version shall be set
+%% {3,4}.
+%% When highest supported version is {3,4} the negotiated version is set to {3,3}.
+effective_version({3,3} , #{versions := [Version|_]}, client) when Version >= {3,4} ->
     Version;
-effective_version(Version, _) ->
+%% Use highest supported version during startup (TLS server, all versions).
+effective_version(undefined, #{versions := [Version|_]}, _) ->
+    Version;
+%% Use negotiated version in all other cases.
+effective_version(Version, _, _) ->
     Version.
+
+
+handle_new_session_ticket(_, #state{ssl_options = #{session_tickets := disabled}}) ->
+    ok;
+handle_new_session_ticket(#new_session_ticket{ticket_nonce = Nonce} = NewSessionTicket,
+                          #state{connection_states = ConnectionStates,
+                                 ssl_options = #{session_tickets := SessionTickets,
+                                                 server_name_indication := SNI},
+                                 connection_env = #connection_env{user_application = {_, User}}})
+  when SessionTickets =:= manual ->
+    #{security_parameters := SecParams} =
+	ssl_record:current_connection_state(ConnectionStates, read),
+    HKDF = SecParams#security_parameters.prf_algorithm,
+    RMS = SecParams#security_parameters.resumption_master_secret,
+    PSK = tls_v1:pre_shared_key(RMS, Nonce, HKDF),
+    send_ticket_data(User, NewSessionTicket, HKDF, SNI, PSK);
+handle_new_session_ticket(#new_session_ticket{ticket_nonce = Nonce} = NewSessionTicket,
+                          #state{connection_states = ConnectionStates,
+                                 ssl_options = #{session_tickets := SessionTickets,
+                                                 server_name_indication := SNI}})
+  when SessionTickets =:= auto ->
+    #{security_parameters := SecParams} =
+	ssl_record:current_connection_state(ConnectionStates, read),
+    HKDF = SecParams#security_parameters.prf_algorithm,
+    RMS = SecParams#security_parameters.resumption_master_secret,
+    PSK = tls_v1:pre_shared_key(RMS, Nonce, HKDF),
+    tls_client_ticket_store:store_ticket(NewSessionTicket, HKDF, SNI, PSK).
+
+
+handle_key_update(#key_update{request_update = update_not_requested}, State0) ->
+    %% Update read key in connection
+    {ok, update_cipher_key(current_read, State0)};
+handle_key_update(#key_update{request_update = update_requested},
+                  #state{protocol_specific = #{sender := Sender}} = State0) ->
+    %% Update read key in connection
+    State1 = update_cipher_key(current_read, State0),
+    %% Send key_update and update sender's write key
+    case send_key_update(Sender, update_not_requested) of
+        ok ->
+            {ok, State1};
+        {error, Reason} ->
+            {error, State1, ?ALERT_REC(?FATAL, ?INTERNAL_ERROR, Reason)}
+    end.
+
+
+update_cipher_key(ConnStateName, #state{connection_states = CS0} = State0) ->
+    CS = update_cipher_key(ConnStateName, CS0),
+    State0#state{connection_states = CS};
+update_cipher_key(ConnStateName, CS0) ->
+    #{security_parameters := SecParams0,
+      cipher_state := CipherState0} = ConnState0 = maps:get(ConnStateName, CS0),
+    HKDF = SecParams0#security_parameters.prf_algorithm,
+    CipherSuite = SecParams0#security_parameters.cipher_suite,
+    ApplicationTrafficSecret0 = SecParams0#security_parameters.application_traffic_secret,
+    ApplicationTrafficSecret = tls_v1:update_traffic_secret(HKDF, ApplicationTrafficSecret0),
+
+    %% Calculate traffic keys
+    #{cipher := Cipher} = ssl_cipher_format:suite_bin_to_map(CipherSuite),
+    {Key, IV} = tls_v1:calculate_traffic_keys(HKDF, Cipher, ApplicationTrafficSecret),
+
+    SecParams = SecParams0#security_parameters{application_traffic_secret = ApplicationTrafficSecret},
+    CipherState = CipherState0#cipher_state{key = Key, iv = IV},
+    ConnState = ConnState0#{security_parameters => SecParams,
+                            cipher_state => CipherState,
+                            sequence_number => 0},
+    CS0#{ConnStateName => ConnState}.
+
+
+send_key_update(Sender, Type) ->
+    KeyUpdate = tls_handshake_1_3:key_update(Type),
+    tls_sender:send_post_handshake(Sender, KeyUpdate).
+
+
+%% Send ticket data to user as opaque binary
+send_ticket_data(User, NewSessionTicket, HKDF, SNI, PSK) ->
+    Timestamp = erlang:system_time(seconds),
+    TicketData = #{hkdf => HKDF,
+                   sni => SNI,
+                   psk => PSK,
+                   timestamp => Timestamp,
+                   ticket => NewSessionTicket},
+    User ! {ssl, session_ticket, {SNI, erlang:term_to_binary(TicketData)}}.
